/****************************************************************************
 *  
 * Project: Olimex AVR ISP with MKII protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: usbscsi.c
 * Description: USB Mass Storage SCSI Module
 * Developer: Dimitar Dimitrov ( dinuxbg __on__ gmail.com )
 *
 * Last change: $Date: 2007-12-12 17:31:49 +0200 (сряда, 12 Декември 2007) $
 * Revision: $Revision: 89 $
 * Id: $Id: usbscsi.c 89 2007-12-12 15:31:49Z dimitar $
 *
 ****************************************************************************/

#include <stdint.h>
#include <avr/pgmspace.h>

#include "usb.h"
#include "usbscsi.h"
#include "usbscsi_conf.h"

#define min(A,B)	((A)<(B) ? (A) : (B))

static enum ms_csw_status sbc_test_unit_ready(const struct sbc_test_unit_ready_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_request_sense(const struct sbc_request_sense_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_format_unit(const struct sbc_format_unit_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_inquiry(const struct sbc_inquiry_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_prevent_allow_medium_removal(const struct sbc_prevent_allow_medium_removal_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_read_capacity(const struct sbc_read_capacity_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_read_10(const struct sbc_read_10_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_verify(const struct sbc_verify_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_write_10(const struct sbc_write_10_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_mode_sense_6(const struct sbc_mode_sense_6_cb *cb, uint32_t *dlen);
static enum ms_csw_status sbc_mode_sense_10(const struct sbc_mode_sense_10_cb *cb, uint32_t *dlen);

/** SCSI SENSE valid bit. */
static uint8_t sbc_sense_valid;

/** SCSI SENSE key code. */
static uint8_t sbc_sense_key;

/** SCSI SENSE pair of additional sense and qualifier codes. */
static uint16_t sbc_sense_addpair;

/** INQUIRY response. */
struct sbc_inquiry_data inquiry PROGMEM = {
	SBC_DEVTYPE_SCSI_DIRECT_ACCESS,		/* device type */

	0,					/* reserved */
	1,					/* removable */
	SBC_VER_NO_STANDARD_CONFORMANCE,	/* version */

	2,					/* response data format */
	0,0,0,0,				/* some weird SCSI options */

	sizeof(struct sbc_inquiry_data)-4,	/* additional length */

	0,0,0,0,0,0,				/* yet another SCSI options */
	0,0,0,0,0,0,0,
	0,

	/* vendor ID */
	{'O','l','i','m','e','x',' ',' '},
	/* product ID */
	{'F','l','a','s','h',' ','D','i','s','k',' ','D','r','i','v','e'},
	/* product revision ID */
	{'0','1','0','0'}
};

/** MODE SENSE response. Includes only FLEXIBLE DISK page. */
struct sbc_mode_param_list mode_sense PROGMEM = {
	/* parameter header */
	{
		(sizeof(struct sbc_mode_param_list)-2)>>8,
		(sizeof(struct sbc_mode_param_list)-2)>>0,
		0,0x80,0,0,
		0,0		/* no block descriptor! */
	},
	/* Mode Page */
	{
		SBC_PAGECODE_FLEXIBLE_DISK,
		sizeof(struct sbc_mode_page)-2,
		0,0,
		4,		/* number of heads */
		32,		/* sectors per track */
		sbclu_block_length()>>8,/* bytes per sector, MSB */
		sbclu_block_length()&0xff,/* bytes per sector, LSB */
		((sbclu_last_lba()+1)/4/32/sbclu_block_length())>>8,/* number of cylinders, MSB */
		((sbclu_last_lba()+1)/4/32/sbclu_block_length())&0xff,/* number of cylinders, LSB */
		{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}
	}
};

/** Clear memory contents. */
static void memclr(void *p, uint8_t size)
{
	uint8_t *pb = (uint8_t*)p;
	while(size) {
		--size;
		*pb++ = 0;
	}
}


/** 
 * Store SCSI sense information for later retreival by the SBC_OP_REQUEST_SENSE
 * command.
 */
static void sbc_set_sense(uint8_t valid, uint8_t key, uint16_t pair)
{
	sbc_sense_valid = valid;
	sbc_sense_key = key;
	sbc_sense_addpair = pair;
}


/** Helper function that sets the SCSI sense to NO SENSE. */
static void sbc_set_sense_no_sense(void)
{
	sbc_set_sense(	1,
			SBC_SENSEKEY_NO_SENSE,
			SBC_ADDSP_NO_ADDITIONAL_SENSE_INFORMATION);
}

/** Helper function that sets the SCSI sense to invalid command parameter. */
static void sbc_set_sense_invalid_param(void)
{
	sbc_set_sense(	1,
			SBC_SENSEKEY_ILLEGAL_REQUEST,
			SBC_ADDSP_INVALID_FIELD_IN_COMMAND_PACKET);
}

/** 
 * Pad the pending IN transfer with @a dlen number of bytes.
 */
static void sbc_pad_in(const uint32_t *dlen)
{
	uint32_t i = *dlen;

	if(!i) return;
	while(i) {
		--i;
		while(!usb_epfifo_ready()) {
			if(!usb_ep_enabled()) return;
		}
		usb_epfifo_push(0x00);
		if(!usb_epfifo_ready()) {
			usb_in_clrbank();
		}
	}
	if(usb_epfifo_bytecount()) usb_in_clrbank();
}

/** 
 * Pad the pending OUT transfer with @a dlen number of bytes.
 */
static void sbc_pad_out(const uint32_t *dlen)
{
	uint32_t i = *dlen;

	if(!i) return;
	while(i) {
		--i;
		while(!usb_out_ready() || !usb_epfifo_ready()) {
			if(!usb_ep_enabled()) return;
		}
		usb_epfifo_pop();
		if(!usb_epfifo_ready()) {
			usb_out_sendack();
			usb_out_clrbank();
		}
	}
	if(usb_epfifo_bytecount()) {
		usb_out_sendack();
		usb_out_clrbank();
	}
}


/** 
 * Send data to USB.
 *
 * @param p Pointer to data.
 * @param size Number of valid data bytes.
 * @param dlen Pointer to the requested number of bytes to transfer. If more
 * than @a size than padding is used. After completion the pointed variable
 * will contain the number of padding bytes.
 * @param is_pgm Whether @a p points to program memory.
 *
 * @return Mass Storage Status Code.
 */
static enum ms_csw_status sbc_send_data(uint8_t *p, uint8_t size, uint32_t *dlen, const uint8_t is_pgm)
{
	/* zero length transfer is not an error */
	if(*dlen) {
		while(*dlen && size) {
			uint8_t c;

			--(*dlen);
			--size;
			while(!usb_epfifo_ready()) {
				if(!usb_ep_enabled()) {
					return MS_CSW_STATUS_PHASE_ERROR;
				}
			}
			if(is_pgm) {
				c = pgm_read_byte(p);
			} else {
				c = *p;
			}
			++p;
			usb_epfifo_push(c);
			if(!usb_epfifo_ready()) {
				usb_in_clrbank();
			}
		}
		if(!*dlen) {
			if(usb_epfifo_bytecount()) usb_in_clrbank();
		} else {
			sbc_pad_in(dlen);
		}
	}

	if(size) {
		sbc_set_sense_invalid_param();
		return MS_CSW_STATUS_PHASE_ERROR;
	} else {
		sbc_set_sense_no_sense();
		return MS_CSW_STATUS_COMMAND_PASSED;
	}
}

/** Handle the case of invalid request parameter. */
static enum ms_csw_status 
sbc_handle_param_error(uint32_t *dlen, const uint8_t dir_d2h) 
{
	if(dir_d2h) {
		sbc_pad_in(dlen);
	} else {
		sbc_pad_out(dlen);
	}
	sbc_set_sense_invalid_param();
	return MS_CSW_STATUS_PHASE_ERROR;
}

/** 
 * Decode a SCSI Block Command and execute the data stage of a USB Mass Storage
 * Bulk-Only transfer.
 *
 * @param cb Pointer to an array, containing a SCSI request and corresponding
 * parameters.
 *
 * @param dlen On entry, must point to a variable, containing the number of
 * bytes for the data stage transfer. On exit it contains the number of 
 * padding bytes.
 *
 * @param dir_d2h Non-zero if data stage direction is Device-To-Host, zero 
 * otherwise.
 *
 * @return Mass Storage Status Code.
 */
enum ms_csw_status 
sbc_decode(const uint8_t cb[16], uint32_t *dlen, const uint8_t dir_d2h)
{
	enum ms_csw_status stat;

	/* define this function macro to save some typing later. */
	#define SBC_EXECCMD(cmd)	sbc_ ## cmd ((const struct sbc_ ## cmd ## _cb *)cb,dlen)

	switch(cb[0]) {
		case SBC_OP_TEST_UNIT_READY:
			if(dir_d2h) {
				sbc_pad_in(dlen);
			} else {
				sbc_pad_out(dlen);
			}
			sbc_set_sense_invalid_param();
			stat = SBC_EXECCMD(test_unit_ready);
			break;
		case SBC_OP_REQUEST_SENSE:
			if(dir_d2h) {
				stat = SBC_EXECCMD(request_sense);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_FORMAT_UNIT:
			if(!*dlen) {
				stat = SBC_EXECCMD(format_unit);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_INQUIRY:
			if(dir_d2h) {
				stat = SBC_EXECCMD(inquiry);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_PREVENT_ALLOW_MEDIUM_REMOVAL:
			if(!*dlen) {
				stat=SBC_EXECCMD(prevent_allow_medium_removal);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_READ_CAPACITY:
			if(dir_d2h) {
				stat = SBC_EXECCMD(read_capacity);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_READ_10:
			if(dir_d2h) {
				stat = SBC_EXECCMD(read_10);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_VERIFY:
			stat = SBC_EXECCMD(verify);
			break;
		case SBC_OP_WRITE_10:
			if(!dir_d2h) {
				stat = SBC_EXECCMD(write_10);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_MODE_SENSE_6:
			if(dir_d2h) {
				stat = SBC_EXECCMD(mode_sense_6);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		case SBC_OP_MODE_SENSE_10:
			if(dir_d2h) {
				stat = SBC_EXECCMD(mode_sense_10);
			} else {
				stat = sbc_handle_param_error(dlen,dir_d2h);
			}
			break;
		default:
			if(dir_d2h) {
				sbc_pad_in(dlen);
			} else {
				sbc_pad_out(dlen);
			}
			sbc_set_sense(1,
				SBC_SENSEKEY_ILLEGAL_REQUEST,
				SBC_ADDSP_NO_ADDITIONAL_SENSE_INFORMATION);
			stat = MS_CSW_STATUS_COMMAND_FAILED;
	}

	return stat;
}


/** Handle the SCSI Test Unit Ready command. */
static enum ms_csw_status 
sbc_test_unit_ready(const struct sbc_test_unit_ready_cb *cb, uint32_t *dlen)
{
	/* we're always ready */
	(void)cb;
	(void)dlen;
	sbc_set_sense_no_sense();
	return MS_CSW_STATUS_COMMAND_PASSED;
}


/** Handle the SCSI Request Sense command. */
static enum ms_csw_status 
sbc_request_sense(const struct sbc_request_sense_cb *cb, uint32_t *dlen)
{
	struct sbc_request_sense_data sense;
	enum ms_csw_status stat;
	uint8_t size;

	memclr(&sense,sizeof(sense));

	size = min(cb->allocation_length,sizeof(sense));
	
	/* fill the structure */
	sense.errcode = SBC_REQSENSE_CURRENT_ERROR;
	sense.valid = sbc_sense_valid;
	sense.sense_key = sbc_sense_key;
	sense.additional_sense_length = sizeof(struct sbc_request_sense_data)-8;
	sense.additional_sense_code = (sbc_sense_addpair>>8) & 0xff;
	sense.additional_sense_code_qualifier = sbc_sense_addpair & 0xff;

	/* send it */
	stat = sbc_send_data((uint8_t*)&sense,size,dlen,0);

	if(cb->allocation_length > sizeof(sense)) {
		sbc_set_sense_invalid_param();
		return MS_CSW_STATUS_PHASE_ERROR;
	} else {
		return stat;
	}
}

/** Handle the SCSI Format Unit command. */
static enum ms_csw_status 
sbc_format_unit(const struct sbc_format_unit_cb *cb, uint32_t *dlen)
{
	(void)dlen;
	
	if((cb->fmtdata==1) && (cb->cmplist==0) && (cb->defect_list_format==7)
			&& (cb->interleave_b1==1) && (cb->interleave_b0==1)) 
	{

		sbclu_format_unit();
		sbc_set_sense_no_sense();
		return MS_CSW_STATUS_COMMAND_PASSED;
	} else {
		sbc_set_sense_invalid_param();
		return MS_CSW_STATUS_COMMAND_FAILED;
	}
}


/** Handle the SCSI Inquiry command. */
static enum ms_csw_status 
sbc_inquiry(const struct sbc_inquiry_cb *cb, uint32_t *dlen)
{
	uint8_t size;

	size = min(cb->allocation_length,sizeof(inquiry));
	
	return sbc_send_data((uint8_t*)&inquiry,size,dlen,1);
}


/** Handle the SCSI Preven Allow Medium Removal command. */
static enum ms_csw_status sbc_prevent_allow_medium_removal(const struct sbc_prevent_allow_medium_removal_cb *cb, uint32_t *dlen)
{
	/* we have nothing to lock */
	(void)cb;
	(void)dlen;
	sbc_set_sense_no_sense();
	return MS_CSW_STATUS_COMMAND_PASSED;
}

/** Handle the SCSI Read Capacity command. */
static enum ms_csw_status 
sbc_read_capacity(const struct sbc_read_capacity_cb *cb, uint32_t *dlen)
{
	struct sbc_read_capacity_data capacity;

	(void)cb;

	/* prepare the answer */
	capacity.last_lba_b3 = (sbclu_last_lba()>>24)&0xff;
	capacity.last_lba_b2 = (sbclu_last_lba()>>16)&0xff;
	capacity.last_lba_b1 = (sbclu_last_lba()>>8)&0xff;
	capacity.last_lba_b0 = (sbclu_last_lba()>>0)&0xff;
	capacity.block_length_b3 = (sbclu_block_length()>>24)&0xff;
	capacity.block_length_b2 = (sbclu_block_length()>>16)&0xff;
	capacity.block_length_b1 = (sbclu_block_length()>>8)&0xff;
	capacity.block_length_b0 = (sbclu_block_length()>>0)&0xff;

	/* send it */
	sbc_set_sense_no_sense();
	return sbc_send_data((uint8_t*)&capacity,sizeof(capacity),dlen,0);
}


/** Handle the SCSI Read 10 command. */
static enum ms_csw_status 
sbc_read_10(const struct sbc_read_10_cb *cb, uint32_t *dlen)
{
	union {
		uint32_t dw;
		uint8_t b[4];
	} lba;

	union {
		uint16_t w;
		uint8_t b[2];
	} len;

#if SBCLU_BLOCK_READ_API
	uint8_t rbuf[USB_EP_MS_IN_MAX_PACKET_SIZE];
#endif
	lba.b[0] = cb->lba_b0;
	lba.b[1] = cb->lba_b1;
	lba.b[2] = cb->lba_b2;
	lba.b[3] = cb->lba_b3;
	len.b[0] = cb->transfer_length_b0;
	len.b[1] = cb->transfer_length_b1;


	if(len.w > *dlen/sbclu_block_length()) {
		sbc_pad_out(dlen);
		return MS_CSW_STATUS_PHASE_ERROR;
	}

	if(!len.w) {
		if(*dlen) sbc_pad_in(dlen);
		return MS_CSW_STATUS_COMMAND_PASSED;
	}

#if SBCLU_BLOCK_READ_API
#else
	sbclu_contread_start(lba.dw);
#endif

	while(len.w) {
		--len.w;
		uint16_t i;

#if SBCLU_BLOCK_READ_API
		sbclu_mem2buf(lba.dw);
#endif
		++lba.dw;
		for(i=0; i<sbclu_block_length(); i += USB_EP_MS_IN_MAX_PACKET_SIZE) {
			while(!usb_epfifo_ready()) {
				if(!usb_ep_enabled()) {
#if SBCLU_BLOCK_READ_API
#else
					sbclu_contread_stop();
#endif
					return MS_CSW_STATUS_PHASE_ERROR;
				}
			}
#if SBCLU_BLOCK_READ_API
			sbclu_rbuf(rbuf,sizeof(rbuf),i);
			for(uint8_t j=0; j<sizeof(rbuf); ++j) {
				usb_epfifo_push(rbuf[j]);
			}
#else
			for(uint8_t j=0; j<USB_EP_MS_IN_MAX_PACKET_SIZE; ++j) {
				usb_epfifo_push(sbclu_contread_fetch());
			}
#endif
			usb_in_clrbank();
		}
		*dlen -= sbclu_block_length();
	}
#if SBCLU_BLOCK_READ_API
#else
	sbclu_contread_stop();
#endif
	if(!*dlen) {
		if(usb_epfifo_bytecount()) usb_in_clrbank();
	} else {
		sbc_pad_in(dlen);
	}

	sbc_set_sense_no_sense();
	return MS_CSW_STATUS_COMMAND_PASSED;
}


/** Handle the SCSI Verify command. */
static enum ms_csw_status 
sbc_verify(const struct sbc_verify_cb *cb, uint32_t *dlen)
{
	union {
		uint32_t dw;
		uint8_t b[4];
	} lba;
	union {
		uint16_t w;
		uint8_t b[2];
	} len;
	enum ms_csw_status stat = MS_CSW_STATUS_COMMAND_PASSED;

#if SBC_READBACK_VERIFY 
	uint8_t vbuf[USB_EP_MS_OUT_MAX_PACKET_SIZE];
#endif /* SBC_READBACK_VERIFY */

	sbc_set_sense_no_sense();

	lba.b[0] = cb->lba_b0;
	lba.b[1] = cb->lba_b1;
	lba.b[2] = cb->lba_b2;
	lba.b[3] = cb->lba_b3;
	len.b[0] = cb->verif_length_b0;
	len.b[1] = cb->verif_length_b1;

	if(cb->bytechk && (len.w > *dlen/sbclu_block_length())) {
		sbc_pad_out(dlen);
		return MS_CSW_STATUS_PHASE_ERROR;
	}

	/* We don't store CRC/ECC information in the FLASH, so we cannot
	 * verify without data to compare. */
	if(!len.w || !cb->bytechk) {
		if(*dlen) sbc_pad_out(dlen);
		return MS_CSW_STATUS_COMMAND_PASSED;
	}

#if SBC_READBACK_VERIFY 
	while(len.w) {
		--len.w;
		uint16_t i;

		sbclu_mem2buf(lba.dw);

		for(i=0; i<sbclu_block_length(); i += sizeof(vbuf)) {
			uint8_t j;

			sbclu_rbuf(vbuf,sizeof(vbuf),i);
			while(!usb_out_ready() || !usb_epfifo_ready()) {
				if(!usb_ep_enabled()) 
					return MS_CSW_STATUS_PHASE_ERROR;
			}
			if(usb_epfifo_bytecount() != sizeof(vbuf)) {
				sbc_set_sense_invalid_param();
				return MS_CSW_STATUS_PHASE_ERROR;
			}
			for(j=0; j<sizeof(vbuf); ++j) {
				if(vbuf[j] != usb_epfifo_pop()) {
					sbc_set_sense(1,
						SBC_SENSEKEY_MISCOMPARE,
						SBC_ADDSP_MISCOMPARE_DURING_VERIFY_OPERATION);
					stat = MS_CSW_STATUS_COMMAND_FAILED;
				}
			}
			usb_out_sendack();
			usb_out_clrbank();
		}
		*dlen -= sbclu_block_length();
		++lba.dw;
	}
#endif	/* SBC_READBACK_VERIFY */

	/* we transmit on packet boundaries here, so no need for usb_in_clrbank */
	sbc_pad_out(dlen);

	return stat;
}

/** Handle the SCSI Write 10 command. */
static enum ms_csw_status 
sbc_write_10(const struct sbc_write_10_cb *cb, uint32_t *dlen)
{

	/*return MS_CSW_STATUS_PHASE_ERROR;*/
	return MS_CSW_STATUS_COMMAND_PASSED;
	/*union {
		uint32_t dw;
		uint8_t b[4];
	} lba;
	union {
		uint16_t w;
		uint8_t b[2];
	} len;
	uint8_t wbuf[USB_EP_MS_OUT_MAX_PACKET_SIZE];

	sbc_set_sense_no_sense();

	lba.b[0] = cb->lba_b0;
	lba.b[1] = cb->lba_b1;
	lba.b[2] = cb->lba_b2;
	lba.b[3] = cb->lba_b3;
	len.b[0] = cb->transfer_length_b0;
	len.b[1] = cb->transfer_length_b1;

	if(len.w > *dlen/sbclu_block_length()) {
		sbc_pad_out(dlen);
		return MS_CSW_STATUS_PHASE_ERROR;
	}
	if(!len.w) {
		if(*dlen) sbc_pad_out(dlen);
		return MS_CSW_STATUS_COMMAND_PASSED;
	}

	while(len.w) {
		--len.w;
		uint16_t i;

		for(i=0; i<sbclu_block_length(); i += sizeof(wbuf)) {
			uint8_t j;

			while(!usb_out_ready() || !usb_epfifo_ready()) {
				if(!usb_ep_enabled()) 
					return MS_CSW_STATUS_PHASE_ERROR;
			}
			if(usb_epfifo_bytecount() != sizeof(wbuf)) {
				sbc_set_sense_invalid_param();
				return MS_CSW_STATUS_PHASE_ERROR;
			}
			for(j=0; j<sizeof(wbuf); ++j) {
				wbuf[j] = usb_epfifo_pop();
			}
			usb_out_sendack();
			usb_out_clrbank();
			sbclu_wbuf(wbuf,sizeof(wbuf),i);
		}
		sbclu_buf2mem(lba.dw);
		*dlen -= sbclu_block_length();
		++lba.dw;
	}
	
	// we transmit on packet boundaries here, so no need for usb_in_clrbank //
	sbc_pad_out(dlen);

	return MS_CSW_STATUS_COMMAND_PASSED;*/
}


/** Handle the SCSI Mode Sense 6 command. */
static enum ms_csw_status 
sbc_mode_sense_6(const struct sbc_mode_sense_6_cb *cb, uint32_t *dlen)
{
	enum ms_csw_status stat;

	stat = sbc_send_data((uint8_t*)&mode_sense,sizeof(mode_sense),dlen,1);

	if((cb->plist_length < sizeof(mode_sense)) || 
			!( (cb->page_code == SBC_PAGECODE_FLEXIBLE_DISK)
			    || (cb->page_code == SBC_PAGECODE_RETURN_ALL) ) )
	{
		sbc_set_sense(1,
			SBC_SENSEKEY_ILLEGAL_REQUEST,
			SBC_ADDSP_INVALID_FIELD_IN_COMMAND_PACKET);
		return MS_CSW_STATUS_COMMAND_FAILED;
	} else {
		sbc_set_sense_no_sense();
		return stat;
	}
}


/** Handle the SCSI Mode Sense 10 command. */
static enum ms_csw_status 
sbc_mode_sense_10(const struct sbc_mode_sense_10_cb *cb, uint32_t *dlen)
{
	enum ms_csw_status stat;

	stat = sbc_send_data((uint8_t*)&mode_sense,sizeof(mode_sense),dlen,1);

	sbc_set_sense_no_sense();
	if((!cb->plist_length_b1 && (cb->plist_length_b0 < sizeof(mode_sense)))
		|| ((cb->page_code != SBC_PAGECODE_FLEXIBLE_DISK)
		    && (cb->page_code != SBC_PAGECODE_RETURN_ALL)))
	{
		sbc_set_sense(1,
			SBC_SENSEKEY_ILLEGAL_REQUEST,
			SBC_ADDSP_INVALID_FIELD_IN_COMMAND_PACKET);
		return MS_CSW_STATUS_COMMAND_FAILED;
	} else {
		return stat;
	}
}

